@@ -1,10 +1,12 @@
 import json
 
+SUPPORTED_OBJECT_TYPES = ['resource', 'alert', 'availability', 'ticket', 'searchpatch', 'resPatchStatus', 'jobData']
 """
 This Class is used for maintaining all the query fields
 """
 class QueryBuilder:
 
+
     def __init__(self, tenant_id, query_params):
         self.tenant_id = tenant_id
         self.object_type = None
@@ -74,10 +76,42 @@ class QueryBuilder:
             query['pageNo'] = self.page_no
         if self.page_size is not None:
             query['pageSize'] = self.page_size
+        query['pageSize'] = self.get_opsql_page_size()
+        return query
+    
+    def get_count_query(self):
+        query = {}
+        if self.object_type is not None:
+            query['objectType'] = self.object_type
+        if self.filter_criteria is not None:
+            query['filterCriteria'] = self.filter_criteria
+        if self.agg_function is not None:
+            query['aggregateFunction'] = 'count'
+        if self.sort_by is not None:
+            query['sortBy'] = self.sort_by
+        if self.sort_order is not None:
+            query['sortByOrder'] = self.sort_order
         return query
 
     def set_page_no(self, page_no):
         self.page_no = page_no
 
     def set_page_size(self, page_size):
-        self.page_size = page_size
\ No newline at end of file
+        self.page_size = page_size
+
+    def get_opsql_page_size(self):
+        page_size = self.page_size
+        if self.fields and self.fields is not None and len(self.fields) > 0:
+            page_size = 500
+            for field in self.fields:
+                if "." in field:
+                    page_size = 100
+                    return page_size
+        if self.filter_criteria and self.filter_criteria is not None and len(self.filter_criteria) > 0:
+            page_size = 500
+            for obj_type in SUPPORTED_OBJECT_TYPES:
+                obj_type = obj_type + "."
+                if obj_type in self.filter_criteria:
+                    page_size = 100
+                    return page_size
+        return page_size
