@@ -12,6 +12,7 @@ from openpyxl import Workbook
 from openpyxl.styles import PatternFill, Font, Alignment
 from openpyxl.worksheet.table import Table, TableStyleInfo
 from openpyxl.utils import get_column_letter
+from openpyxl.chart.label import DataLabelList
 from openpyxl.chart import (
     DoughnutChart,
     ScatterChart,
@@ -19,8 +20,13 @@ from openpyxl.chart import (
     Reference,
     Series,
     LineChart,
+    PieChart
 )
 from openpyxl.chart.series import DataPoint
+from openpyxl.cell.cell import ILLEGAL_CHARACTERS_RE
+import re
+from decimal import Decimal
+from fractions import Fraction
 
 from analytics_sdk.utilities import (
     BASE_API_URL,
@@ -30,9 +36,8 @@ from analytics_sdk.utilities import (
     API_RETRY,
     API_TIME_STACKS,
     APP_DISPLAY_NAME
-    )
+)
 
-#app_name = os.getenv("PLATFORM_ROUTE")
 chart_colors = ["0077C8", "e35420", "32a3df", "673ab7", "9c27b0", "F5BE0A","e30255","a002e3","0079e3","05e3a8","f5d255","f36686","c661e9","24a5ed","62ecbf","F7E487","fb9fb5","e39df0","79ccef","91F4D5","f8ba7e","f3a5ea","cd91eb","87E4F0","cdf38d","f78f4a","eb6dd6","9352e5","5ad1e9","afeb5a","f55a22","e302be","6834E3","02bee3","8ae305"]
 
 class ExcelRenderer:
@@ -50,7 +55,7 @@ class ExcelRenderer:
         self.report_gen_start_time = report_gen_start_time
         self.report_gen_completed_time = report_gen_completed_time
         self.additional_props = additional_props
-        
+
         def get_token():
             headers = {"Content-Type" : "application/x-www-form-urlencoded" , "Accept" : "application/json"};
             post_data = {"grant_type": "client_credentials", "client_id" : API_KEY, "client_secret" : API_SECRET};
@@ -63,10 +68,10 @@ class ExcelRenderer:
 
         def get_headers():
             headers = {
-                            'Content-Type': 'application/json',
-                            'Accept': 'application/json',
-                            'Authorization': f'Bearer {get_token()}'
-                        }
+                'Content-Type': 'application/json',
+                'Accept': 'application/json',
+                'Authorization': f'Bearer {get_token()}'
+            }
 
             return headers
 
@@ -94,7 +99,7 @@ class ExcelRenderer:
         def diff_sec(st, et):
             difference = int(et - st)
             return difference
-        
+
         url = BASE_API_URL + f'/reporting/api/v3/tenants/{tenantId}/runs/{analysis_run}/summary'
         api_proce_before_time = time.time()
         res = call_requests('GET', url)
@@ -110,8 +115,22 @@ class ExcelRenderer:
 
         self.res = res
 
+    def get_value(self, value):
+        if value:
+            try:
+                value = float(value)
+                return value
+            except ValueError:
+                return value
+        return value
+    
+
     def _load_data(self, ws, data, row_start, col_start, adjust_column=True, x_axis_date_format=None):
         column_widths = {}
+        if isinstance(data, list):
+            if len(data) >= 1000005:
+                data = data[0:1000005]
+
         for row_delta, row in enumerate(data):
             for col_delta, val in enumerate(row):
                 col = col_start + col_delta
@@ -119,13 +138,24 @@ class ExcelRenderer:
                 if x_axis_date_format and row_delta > 0 and col_delta == 0:
                     cell.value = datetime.strptime(val, x_axis_date_format).date()
                 else:
-                    cell.value = val
-
+                    try:
+                        if APP_ID == 'AVAILABILITY-DETAILS':
+                            if isinstance(val, str) and 'COLOR-CODE_' in val:
+                                vals = val.split('_') #COLOR-CODE_{color_code}_{value}
+                                if len(vals) == 3:
+                                    val = vals[2]
+                                    if vals[1] is not None and len(vals[1]):
+                                        cell.fill = PatternFill("solid", fgColor=vals[1])
+                        cell.value = self.get_value(val)
+                    except Exception as e:
+                        if isinstance(val, str):
+                            cell.value = ILLEGAL_CHARACTERS_RE.sub(r'', val)
+                        else:
+                            cell.value = self.get_value(val)
                 if col in column_widths:
                     column_widths[col] = max(len(str(val)), column_widths[col])
                 else:
                     column_widths[col] = max(len(str(val)), ws.column_dimensions[get_column_letter(col)].width)
-
         # handle cell width
         if adjust_column:
             for col, column_width in column_widths.items():
@@ -148,6 +178,61 @@ class ExcelRenderer:
         table = Table(displayName=table_name, ref=ref, tableStyleInfo=TableStyleInfo(name="TableStyleMedium9"))
         ws.add_table(table)
 
+
+    def add_metric_table(self, ws, table_data):
+        """
+        add metric table component
+        """
+        row_start = ExcelRenderer.ROW_START + table_data['start-row'] + 1
+        col_start = table_data['start-col']
+        self._load_data(ws, table_data['data'], row_start, col_start)
+        row_span = len(table_data['data'])
+        col_span = len(table_data['data'][0])
+        if 'merge_cells' in table_data:
+            for merge_cell in table_data['merge_cells']:
+                ws.merge_cells(merge_cell)
+        fill_color = 'eeeeee'
+        if 'cell_color' in table_data:
+            fill_color = table_data['cell_color']
+        if 'color_cells' in table_data:
+            min_row = table_data['color_cells'][0]
+            max_row = table_data['color_cells'][1]
+            min_col = table_data['color_cells'][2]
+            max_col = table_data['color_cells'][3]
+
+            for rows in ws.iter_rows(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col):
+                for cell in rows:
+                    cell.fill = PatternFill("solid", fgColor=fill_color)
+                    cell.alignment = Alignment(horizontal="center")
+
+        div_color = 'dee6ef'
+        col_min_col = 13
+        if 'divider-color' in table_data:
+            div_color = table_data['divider-color']
+        if 'col_min_col' in table_data:
+            col_min_col = table_data['col_min_col']
+        if 'col_divider' in table_data:
+            if table_data['col_divider'] and table_data['col_divider'] is not None:
+                col_min_row = table_data['col_divider'][0]
+                col_max_row = table_data['col_divider'][1]
+                col_length = table_data['col_divider'][2]
+                count = 0
+                for i in range(col_length):
+                    if count > 0:
+                        col_min_col = col_min_col+4
+                    for rows in ws.iter_rows(min_row=col_min_row, max_row=col_max_row, min_col=col_min_col, max_col=col_min_col):
+                        for cell in rows:
+                            cell.fill = PatternFill(start_color=div_color, end_color=div_color, fill_type='solid')
+                    count+=1
+
+        ref = f'{get_column_letter(col_start)}{row_start}:{get_column_letter(col_start+col_span-1)}{row_start+row_span-1}'
+        ts = time.time()
+        table_name = f'Table{random.randint(0, 100)}_{ts}'
+
+        table = Table(displayName=table_name, ref=ref, tableStyleInfo=TableStyleInfo(name="TableStyleMedium9"))
+        ws.add_table(table)
+
+
     def add_doughnut_chart(self, ws, chart_data):
         """
         add doughnut chart component
@@ -157,7 +242,11 @@ class ExcelRenderer:
         self._load_data(ws, chart_data['data'], row_start, col_start, False)
         row_span = len(chart_data['data'])
 
-        chart = DoughnutChart(holeSize=50)
+        hole_size = 50
+        if 'hole-size' in chart_data:
+            hole_size = chart_data.get('hole-size')
+
+        chart = DoughnutChart(holeSize=hole_size)
         labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
         data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
         chart.add_data(data, titles_from_data=True)
@@ -167,15 +256,71 @@ class ExcelRenderer:
         chart.width = chart_data.get('width', 11)
         chart.height = chart_data.get('height', 5)
 
+        ##Set the DataLabelList to show percentage of pie chart labels.
+        #chart.dataLabels = DataLabelList()
+        #chart.dataLabels.showVal = False
+        #chart.dataLabels.showCatName = False
+        #chart.dataLabels.showLegendKey = False
+
         slices = [DataPoint(idx=i) for i in range(row_span-1)]
         chart.series[0].data_points = slices
-        colors = ["0077C8", "32a3df", "673ab7", "9c27b0"]
+
+        #colors = ["0077C8", "32a3df", "673ab7", "9c27b0"]
+        chart_colors = self.generate_pie_chart_colors(len(slices))
+        colors = chart_colors
+
+        if 'colors' in chart_data and chart_data['colors']:
+            colors = chart_data['colors']+colors
 
         for idx, slice in enumerate(slices):
-            slice.graphicalProperties.solidFill = colors[idx % 4]
+            #slice.graphicalProperties.solidFill = colors[idx % 4]
+            slice.graphicalProperties.solidFill = colors[idx % len(colors)]
 
         ws.add_chart(chart, chart_data['chart-position'])
 
+
+    def add_pie_chart(self, ws, chart_data):
+        """
+        add doughnut chart component
+        """
+        row_start = ExcelRenderer.ROW_START + chart_data['start-row']
+        col_start = chart_data['start-col']
+        self._load_data(ws, chart_data['data'], row_start, col_start, False)
+        row_span = len(chart_data['data'])
+
+        chart = PieChart()
+        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
+        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
+        chart.add_data(data, titles_from_data=True)
+        chart.set_categories(labels)
+        chart.title = chart_data['chart-title']
+        chart.style = 12
+        chart.width = chart_data.get('width', 11)
+        chart.height = chart_data.get('height', 5)
+
+        ##Set the DataLabelList to show percentage of pie chart labels.
+        #chart.dataLabels = DataLabelList()
+        #chart.dataLabels.showVal = False
+        #chart.dataLabels.showCatName = False
+        #chart.dataLabels.showLegendKey = False
+
+        slices = [DataPoint(idx=i) for i in range(row_span-1)]
+        chart.series[0].data_points = slices
+
+        #colors = ["0077C8", "32a3df", "673ab7", "9c27b0"]
+        chart_colors = self.generate_pie_chart_colors(len(slices))
+        colors = chart_colors
+
+        if 'colors' in chart_data and chart_data['colors']:
+            colors = chart_data['colors']+colors
+
+        for idx, slice in enumerate(slices):
+            #slice.graphicalProperties.solidFill = colors[idx % 4]
+            slice.graphicalProperties.solidFill = colors[idx % len(colors)]
+
+        ws.add_chart(chart, chart_data['chart-position'])
+
+
     def add_bar_chart(self, ws, chart_data):
         """
         add bar chart component
@@ -205,12 +350,25 @@ class ExcelRenderer:
         s1.graphicalProperties.line.solidFill = "0077C8"
         s1.graphicalProperties.solidFill = "0077C8"
 
+        multicolor_check = False
+        multi_colors = chart_colors
+        if 'is-multicolored-chart' in chart_data:
+            multicolor_check = chart_data['is-multicolored-chart']
+        if 'colors' in chart_data and chart_data['colors']:
+            multi_colors = chart_data['colors']
+
         for idx, s in enumerate(chart.series):
-            index = len(chart_colors)%(idx+1)
-            if index >= len(chart_colors):
-                index = 0
-            s.graphicalProperties.line.solidFill = chart_colors[index]
-            s.graphicalProperties.solidFill = chart_colors[index]
+            if multicolor_check:
+                for i in range(len(multi_colors)):
+                    pt = DataPoint(idx=i)
+                    pt.graphicalProperties.solidFill = multi_colors[i]
+                    s.dPt.append(pt)
+            else:
+                index = len(chart_colors)%(idx+1)
+                if index >= len(chart_colors):
+                    index = 0
+                s.graphicalProperties.line.solidFill = chart_colors[index]
+                s.graphicalProperties.solidFill = chart_colors[index]
 
         ws.add_chart(chart, chart_data['chart-position'])
 
@@ -247,12 +405,20 @@ class ExcelRenderer:
         s1.graphicalProperties.line.solidFill = "0077C8"
         s1.graphicalProperties.solidFill = "0077C8"
 
+        add_chart_colors = chart_colors
+        colors_check = False
+        if 'colors' in chart_data and chart_data['colors']:
+            add_chart_colors = chart_data['colors']+add_chart_colors
+            colors_check = True
+
         for idx, s in enumerate(chart.series):
-            index = len(chart_colors)%(idx+1)
-            if index >= len(chart_colors):
+            index = idx
+            if not colors_check:
+                index = len(add_chart_colors)%(idx+1)
+            if index >= len(add_chart_colors):
                 index = 0
-            s.graphicalProperties.line.solidFill = chart_colors[index]
-            s.graphicalProperties.solidFill = chart_colors[index]
+            s.graphicalProperties.line.solidFill = add_chart_colors[index]
+            s.graphicalProperties.solidFill = add_chart_colors[index]
 
         ws.add_chart(chart, chart_data['chart-position'])
 
@@ -308,6 +474,8 @@ class ExcelRenderer:
         if not no_of_lines:
             no_of_lines = 1
         colors = ["0077C8", "e35420", "32a3df", "673ab7", "9c27b0"]
+        if 'colors' in chart_data and chart_data['colors']:
+            colors = chart_data['colors']+colors
 
         chart = LineChart()
         chart.style = 5
@@ -326,9 +494,9 @@ class ExcelRenderer:
             chart.add_data(data, titles_from_data=True)
             chart.set_categories(labels)
             series = chart.series[i]
-            series.marker.graphicalProperties.solidFill = colors[i % 5]  # Marker filling
-            series.marker.graphicalProperties.line.solidFill = colors[i % 5]  # Marker outline
-            series.graphicalProperties.line.solidFill = colors[i % 5]
+            series.marker.graphicalProperties.solidFill = colors[i % no_of_lines]  # Marker filling
+            series.marker.graphicalProperties.line.solidFill = colors[i % no_of_lines]  # Marker outline
+            series.graphicalProperties.line.solidFill = colors[i % no_of_lines]
             series.graphicalProperties.line.width = 24050  # width in EMUs
             series.smooth=True        # Draw a smooth line
             if x_axis_date_format:
@@ -336,8 +504,64 @@ class ExcelRenderer:
                 chart.x_axis.majorTimeUnit = "days"
 
         ws.add_chart(chart, chart_data['chart-position'])
-        
-        
+
+
+    def add_bar_line_trend_chart(self, ws, chart_data):
+        """
+        add bar line trend chart component
+        """
+        row_start = ExcelRenderer.ROW_START + chart_data['start-row']
+        col_start = chart_data['start-col']
+        self._load_data(ws, chart_data['data'], row_start, col_start, False)
+        row_span = len(chart_data['data'])
+
+        chart = BarChart()
+        chart.type = "col"
+        chart.style = 10
+        chart.title = chart_data['chart-title']
+        chart.x_axis.title = chart_data.get('x-axis-title')
+        chart.y_axis.title = chart_data.get('y-axis-title')
+
+        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
+        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
+        chart.add_data(data, titles_from_data=True)
+        chart.set_categories(labels)
+        chart.style = 4
+        chart.legend = None
+        chart.width = chart_data.get('width', 17)
+        chart.height = chart_data.get('height', 8.5)
+
+        chart2 = LineChart()
+        #chart2.type = "col"
+        chart2.style = 10
+        chart2.title = chart_data['chart-title']
+        chart2.x_axis.title = chart_data.get('x-axis-title')
+        chart2.y_axis.title = chart_data.get('y-axis-title')
+
+        labels = Reference(ws, min_col=col_start, min_row=row_start+1, max_row=row_start+row_span-1)
+        data = Reference(ws, min_col=col_start+1, min_row=row_start, max_row=row_start+row_span-1)
+        chart2.add_data(data, titles_from_data=True)
+        chart2.set_categories(labels)
+        chart2.style = 4
+        chart2.legend = None
+        chart2.width = chart_data.get('width', 17)
+        chart2.height = chart_data.get('height', 8.5)
+
+        s1 = chart.series[0]
+        s1.graphicalProperties.line.solidFill = "0077C8"
+        s1.graphicalProperties.solidFill = "0077C8"
+
+        for idx, s in enumerate(chart.series):
+            index = len(chart_colors)%(idx+1)
+            if index >= len(chart_colors):
+                index = 0
+            s.graphicalProperties.line.solidFill = chart_colors[index]
+            s.graphicalProperties.solidFill = chart_colors[index]
+
+        chart += chart2
+        ws.add_chart(chart, chart_data['chart-position'])
+
+
     def client_names_ids(self, customer_name):
         names = ''
         client_id = ''
@@ -359,7 +583,10 @@ class ExcelRenderer:
         """
         add glossary sheet
         """
-        for row_idx in range(1, 30):
+        data_len = 35
+        if sheet_resp is not None and 'data' in sheet_resp and sheet_resp['data'] is not None and len(sheet_resp['data']) >= data_len:
+            data_len = len(sheet_resp['data'])+5
+        for row_idx in range(1, data_len):
             row = ws.row_dimensions[row_idx]
             row.fill = PatternFill("solid", fgColor="eeeeee")
 
@@ -400,7 +627,7 @@ class ExcelRenderer:
 
         ws.column_dimensions['B'].width = 24
         ws.column_dimensions['C'].width = 24
-        
+
         c_name = ''
         c_id = ''
         opsql_params = None
@@ -413,7 +640,7 @@ class ExcelRenderer:
 
             if 'opsql_query' in additional_props:
                 opsql_params = additional_props['opsql_query']
-        
+
         cell = ws['B4']
         cell.value = 'App'
         cell.font = Font(color="00598B")
@@ -475,7 +702,7 @@ class ExcelRenderer:
         first_name = res.json()['analysisRun']['createdBy']['firstName']
         last_name = res.json()['analysisRun']['createdBy']['lastName']
         login_name = res.json()['analysisRun']['createdBy']['loginName']
-        cell.value = f'{first_name} {last_name}, {login_name}'
+        cell.value = f'{first_name} {last_name} ({login_name})'
         cell.font = Font(color="000000")
         cell.fill = PatternFill("solid", fgColor="eeeeee")
 
@@ -551,6 +778,17 @@ class ExcelRenderer:
                 return all_values
         return value
 
+
+    def generate_pie_chart_colors(self, num_colors):
+        colors=["0077C8", "00A3E0", "673AB7", "9C27B0", "E91E63", "F47925"]
+        random.seed(123)  # 123 Fixed seed value
+        #colors = []
+        for _ in range(num_colors):
+            hex_color = '{:06x}'.format(random.randint(0, 0xFFFFFF))
+            colors.append(hex_color)
+        return colors
+
+
     def add_header(self, ws, sheet_data):
         """
         add header to the sheet
@@ -561,7 +799,7 @@ class ExcelRenderer:
 
         ws.merge_cells('B1:Z2')
         cell = ws['B1']
-        cell.value = APP_ID.replace('-', ' ').title()
+        cell.value = APP_DISPLAY_NAME.replace('-', ' ').title()
         cell.font = Font(color="00598B", bold=True)
         cell.fill = PatternFill("solid", fgColor="eeeeee")
         cell.alignment = Alignment(vertical="center")
@@ -576,21 +814,37 @@ class ExcelRenderer:
         cell.font = Font(color=color, bold=True)
 
     def add_component(self, ws, component_data):
+
+        if component_data and 'data' in component_data:
+            data = component_data['data']
+            if isinstance(data, list):
+                if len(data) >= 1000005:
+                    logger.error('Data exceeding 1Million records.. Excluding excessive records..')
+                    data = data[0:1000005]
+                    component_data['data'] = data
+
         if component_data['type'] == 'table':
             if component_data.get('title'):
                 self.add_component_title(ws, component_data)
-            self.add_table(ws, component_data)
+            if 'metric_sheet' in component_data:
+                self.add_metric_table(ws, component_data)
+            else:
+                self.add_table(ws, component_data)
         elif component_data['type'] == 'name':
             if component_data.get('title'):
                 self.add_component_title(ws, component_data)
         elif component_data['type'] == 'doughnut-chart':
             self.add_doughnut_chart(ws, component_data)
+        elif component_data['type'] == 'pie-chart':
+            self.add_pie_chart(ws, component_data)
         elif component_data['type'] == 'bar-chart':
             self.add_bar_chart(ws, component_data)
         elif component_data['type'] == 'scatter-chart':
             self.add_scatter_chart(ws, component_data)
         elif component_data['type'] == 'line-chart':
             self.add_line_chart(ws, component_data)
+        elif component_data['type'] == 'bar-line-trend-chart':
+            self.add_bar_line_trend_chart(ws, component_data)
         elif component_data['type'] == 'stack-bar-chart':
             self.add_stack_bar_chart(ws, component_data)
 
