@@ -2,10 +2,12 @@ import os
 import time
 import datetime
 import logging
+import random
 from io import BytesIO
-from datetime import datetime
 import json
 import traceback
+from datetime import datetime
+import pytz
 
 from urllib.request import urlopen
 from email.mime.text import MIMEText
@@ -13,13 +15,12 @@ from email.mime.application import MIMEApplication
 from email.mime.multipart import MIMEMultipart
 from flask.wrappers import Response
 
-import jwt
 import boto3
 import botocore
 from botocore.exceptions import NoCredentialsError
 import flask
 import requests
-
+from flask import request
 from .constants import DATETIME_FORMAT
 logger = logging.getLogger(__name__)
 
@@ -45,6 +46,9 @@ if storage_name == 's3':
     SECRET_ACCESS_KEY = os.getenv('SECRET_ACCESS_KEY')
     REGION_NAME = os.getenv('REGION_NAME')
     BUCKET_NAME = os.getenv('BUCKET_NAME')
+    S3_ENDPOINT_URL = os.getenv('S3_ENDPOINT_URL', None) # This Property used for on-prime pods to connect cloud
+    if S3_ENDPOINT_URL is not None and 'https://' not in S3_ENDPOINT_URL:
+        S3_ENDPOINT_URL = 'https://'+S3_ENDPOINT_URL
 elif storage_name == 'ceph':
     ACCESS_KEY_ID = os.getenv('ACCESS_KEY_ID')
     SECRET_ACCESS_KEY = os.getenv('SECRET_ACCESS_KEY')
@@ -116,6 +120,8 @@ def call_requests(method, url, params=None, data=None, json=None, verify=True):
     while retry <= API_RETRY:
         try:
             resp = requests.request(method, url, params=params, data=data, json=json, headers=ACCESS_HEADER, verify=verify)
+            logger.info(f'Params :: {params}  Data :: {data}')
+            logger.info(f'Response = {resp}')
             if resp.status_code == 407:
                 if TOKEN_RETRIES < 5:
                     logging.info(f'Token expired, re-generating token..{TOKEN_RETRIES}' )
@@ -146,25 +152,20 @@ def login_call_requests(method, url, params=None, data=None, json=None, verify=T
     retry = 1
     resp = None
     while retry <= API_RETRY:
-        logger.info(f"Processing authentication request at {retry} iteration ")
         print(f"Processing authentication request at {retry} iteration ")
         try:
             resp = requests.request(method, url, params=params, data=data, json=json, headers=headers, verify=verify)
-            logger.info("authentication response is %s %s", resp,url)
             print("authentication response is ", resp, url)
             if not resp.ok:
                 time.sleep(retry * 2)
                 retry+=1
-                logger.info(f"Invalid response, going to {retry} iteration ")
                 print(f"Invalid response, going to {retry} iteration ")
                 continue
         except requests.exceptions.ConnectionError:
             time.sleep(retry * 2)
             retry+=1
-            logger.info(f"Got the exception, going to {retry} iteration ")
             print(f"Got the exception, going to {retry} iteration ")
             continue
-        logger.info("Response is %s  ", resp)
         print("Response is ", resp)
         return resp
     return resp
@@ -188,18 +189,17 @@ def login_call_get_requests(url, params=None, verify=True):
 
 def is_authenticated():
     REQUIRE_AUTH_REDIRECT = os.getenv('REQUIRE_AUTH_REDIRECT') == 'true'
-    logger.info("Getting authentication configuration value %s", REQUIRE_AUTH_REDIRECT)
-    print("Getting authentication configuration value ", REQUIRE_AUTH_REDIRECT)
+    print("Getting authentication configuration value is ", REQUIRE_AUTH_REDIRECT)
     if not REQUIRE_AUTH_REDIRECT:
-        logger.info("authentication is not verifying,means bypassing authentication")
         print("authentication is not verifying,means bypassing authentication")
         return True
-    
-    url = f'{BASE_API_URL}/api/v2/users/me'
-    logger.info("authentication is verifying, request URL is %s",url)
-    print("authentication is verifying, request URL is ",url)
-    res = login_call_get_requests(url)
-    return res.status_code == 200
+    if get_jwt_token():
+        url = f'{BASE_API_URL}/api/v2/users/me'
+        print("authentication is verifying, request URL is ",url)
+        res = login_call_get_requests(url)
+        return res.status_code == 200
+    print("Invalid authentication, redirecting to login url")
+    return False
 
 
 
@@ -243,8 +243,9 @@ def get_result_by_run(run_id, field=None, default_value=None):
             return result
         else:
             logger.info("No storages are found")
-    except Exception:
+    except Exception as e:
         logger.error('An error occurred (NoSuchKey) when calling the GetObject operation: The specified key does not exist')
+        traceback.print_stack()
         pass
 
 
@@ -269,7 +270,9 @@ def get_s3_client():
     return boto3.client('s3',
                         region_name=REGION_NAME,
                         aws_access_key_id=ACCESS_KEY_ID,
-                        aws_secret_access_key=SECRET_ACCESS_KEY)
+                        aws_secret_access_key=SECRET_ACCESS_KEY,
+                        endpoint_url=S3_ENDPOINT_URL
+                        )
     
 
 def get_ceph_resource():
@@ -323,7 +326,8 @@ def upload_to_storage(content, location):
         s3 = boto3.resource('s3',
                             region_name=REGION_NAME,
                             aws_access_key_id=ACCESS_KEY_ID,
-                            aws_secret_access_key=SECRET_ACCESS_KEY)
+                            aws_secret_access_key=SECRET_ACCESS_KEY,
+                            endpoint_url=S3_ENDPOINT_URL)
         object_url = f'https://{BUCKET_NAME}.s3.{REGION_NAME}.amazonaws.com/{location}'
         try:
             s3.Bucket(BUCKET_NAME).put_object(Body=content,
@@ -417,7 +421,7 @@ def generate_pdf(analysis_run):
         raise Exception(f'Generate_pdf:: pdf generation failed after max tries ({API_RETRY}), for run ::: {pdf.analysis_run}')
     except Exception as e:
         traceback.print_exc()
-        err_msg = f'Generate_pdf:: Exeception - pdf generation failed after max tries({API_RETRY}), for run ::: {analysis_run}'
+        err_msg = f'Generate_pdf:: Exeception - pdf generation failed after max tries({API_RETRY}), for run ::: {pdf.analysis_run}'
         raise Exception(err_msg)
 
 
@@ -460,6 +464,30 @@ def diff_sec(st, et):
     return difference
 
 
+def add_custom_headers(response):
+    # Add CORS headers
+    response.headers['Access-Control-Max-Age'] = 3600
+    response.headers['Access-Control-Expose-Headers'] = 'Content-Length'
+    response.headers['Access-Control-Allow-Headers'] = 'Range'
+    response.headers['Access-Control-Allow-Methods'] = 'POST, GET, DELETE, PATCH'
+    response.headers['X-Frame-Options'] = 'SAMEORIGIN'
+    response.headers['Content-Security-Policy'] = "script-src: 'self' 'unsafe-inline' https://www.google.com"
+    response.headers['Strict-Transport-Security'] = "max-age=31536000 ; includeSubDomains;preload;"
+    response.headers['X-Content-Type-Options'] = 'nosniff'
+    response.headers['X-Xss-Protection'] = "1;mode=block"
+    return response
+
+
+# Cache Control Mechanism
+def add_cache_control(response):
+    # Check if the requested URL matches the _dash-component-suites route
+    path = request.path
+    if "/opsramp-analytics-utils/" in path or path.endswith("/_dash-layout") or path.endswith("/_dash-dependencies"):
+        # Add Cache-Control header to the response for browser caching
+        response.headers['Cache-Control'] = 'max-age=31536000'  # Cache for 365 days
+    return response
+
+
 def update_status_url(analysisRunId, tenantId, genStime, genEtime, status=None):
 
     url = BASE_API_URL + f'/reporting/api/v3/tenants/{tenantId}/runs/{analysisRunId}'
@@ -505,6 +533,20 @@ def update_results_url(gen_start_time, analysisRunId, tenantId, json_result_url=
     if api_proce_diff > API_TIME_STACKS:
         logging.info('Database update response took %d (greater than %d) seconds', api_proce_diff, API_TIME_STACKS)
     logger.info('Database update response is %s', res)
+    
+    
+def upload_file(run_id, reportname, filepath):
+    if storage_name == 's3':
+        excel_file_location = f'{PLATFORM_ROUTE}/{run_id}/xls/' + reportname
+        excel_url = upload_excel_s3(filepath, BUCKET_NAME, excel_file_location)
+    elif storage_name == 'ceph':
+        excel_file_location = f'{APP_ID.lower()}/{run_id}/xls/' + reportname
+        excel_url = upload_excel_ceph(filepath, BUCKET_NAME, excel_file_location)
+    else:
+        logger.info("No storages are found")
+    
+    delete_excel_file(filepath)
+    return excel_url
 
 
 #Upload excel file to s3
@@ -604,7 +646,6 @@ def init_mail(org_id, run_id):
     logger.info('finished init_mail method')
 
 
-
 def send_internal_failure_notification_mail(run_id, app_name, err_msg, partner_id, partner_name, client_id, client_name):
     logger.info('Entered into send_internal_failure_notification_mail method')
     try:
@@ -632,6 +673,24 @@ def send_internal_failure_notification_mail(run_id, app_name, err_msg, partner_i
     logger.info('finished send_internal_failure_notification_mail method')
 
 
+def send_user_failure_notification(run_id, org_id):
+    logger.info('Entered into send_user_failure_notification method')
+    try:
+        url = BASE_API_URL + f'/reporting/api/v3/tenants/{org_id}/runs/{run_id}/failureNotification'
+        data = {}
+        t1 = int(time.time())
+        res = call_post_requests(url , data=json.dumps(data), verify=False)
+        t2 = int(time.time())
+        duration = t2-t1
+        if duration > API_TIME_STACKS:
+            logging.info('Send internal failure notification to customer mail response took %d (greater than %d) seconds', duration, API_TIME_STACKS)
+        logger.info('Send internal failure notification to customer mail response is %s', res)
+    except Exception as e:
+        raise e
+        logger.error("Exception raised due to : " + repr(e))
+        traceback.print_exc()
+    logger.info('finished send_user_failure_notification method')
+    
 
 def upload_excel(analysis_run, excel_file):
     timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
@@ -705,51 +764,98 @@ def get_last_run(summary_res):
 def get_app_description(summary_res):
     return summary_res['app_description']
 
+def prepare_page_size_page_no(data, page_no, page_size):
+    if data and data is not None:
+        if "pageNo" in data:
+            data["pageNo"] = page_no
+        if "pageSize" not in data:
+            data["pageSize"] = page_size
+    return data
+
+
+def add_page_no_page_size_for_url(add_page_no, url, page_no, page_size):
+    uri = url
+    if add_page_no:
+        if "pageSize" not in url:
+            uri = url + f'&pageNo={page_no}&pageSize={page_size}'
+        else:
+            uri = url + f'&pageNo={page_no}'
+    return uri
+
 
 def get_paginated_api_results(method, url, data, type, add_pageNo=True):
     resp = []
     page_no=1
     nextPage= True
-    pageSize= 100
+    page_size= 100
 
     try:
+        error_message = ''
         while (nextPage != False):
             data = json.loads(data)
-            if "pageNo" in data:
-                data["pageNo"] = page_no
-                data["pageSize"] = pageSize
+            data = prepare_page_size_page_no(data, page_no, page_size)
             data=json.dumps(data)
             if method == 'POST':
                 res = get_post_request_results(url, data, type)
             else:
-                if add_pageNo:
-                    uri = url + f'&pageNo={page_no}&pageSize={pageSize}'
-                else:
-                    uri = url
+                uri = add_page_no_page_size_for_url(add_pageNo, url, page_no, page_size)
                 res = get_response(uri, type)
+            if res == None or not res.ok or (res.json() is not None and "results" not in res.json()) or (res is not None and "nextPage" not in res.json()):
+                error_message = f'Get {type} API is failed, url ::: {url}, response ::: {res}'
+                logger.error('Get %s API is failed, url ::: %s, response ::: %s ', type, url, res)
+                retry_count = 1
+                while (retry_count <= API_RETRY):
+                    data = json.loads(data)
+                    data = prepare_page_size_page_no(data, page_no, page_size)
+                    data=json.dumps(data)
+
+                    time.sleep(1)
+                    retry_count += 1
+                    if method == 'POST':
+                        res = get_post_request_results(url, data, type)
+                    else:
+                        uri = add_page_no_page_size_for_url(add_pageNo, url, page_no, page_size)
+                        res = get_response(uri, type)
+                    if res == None or not res.ok:
+                        error_message = f'Get {type} API is failed, url ::: {url}, response ::: {res}, response_json ::: {res.json()}, API_Retry_Count ::: {retry_count}'
+                        logger.error('Get %s API is failed, url ::: %s, response ::: %s, response_json ::: %s, API_Retry_Count ::: %s ', type, url, res, res.json(), retry_count)
+                        res = None
+                        if retry_count > 3:
+                            raise Exception(error_message)
+                    elif res.json() is not None and "results" not in res.json():
+                        error_message = f"Get {type} results keyword is missing in API response, url ::: {url}, response ::: {res}, response_json ::: {res.json()}, API_Retry_Count ::: {retry_count}"
+                        logger.error('Get %s results keyword is missing in API response, url ::: %s, response : %s, response_json ::: %s, API_Retry_Count ::: %s ', type, url, res, res.json(), retry_count)
+                        res = None
+                        if retry_count > 3:
+                            raise Exception(error_message)
+                    elif res.json() is not None and "nextPage" not in res.json():
+                        error_message = f"Get {type} nextPage keyword is missing in API response, url : {url}, response ::: {res}, response_json ::: {res.json()}, API_Retry_Count ::: {retry_count}"
+                        logger.error('Get %s nextPage keyword is missing in API response, url ::: %s, response : %s, response_json ::: %s, API_Retry_Count ::: %s ', type, url, res, res.json(), retry_count)
+                        res = None
+                        if retry_count > 3:
+                            raise Exception(error_message)
+                    else:
+                        break
+                if res == None or "results" not in res.json() or len(res.json()['results'])==0:
+                    logger.error('After retrying for %s times, Get %s API results are empty, url is %s', API_RETRY, type, url)
+                    return None
+            #else:
+            result = res.json()['results']
+            resp.append(result)
+
+            if "nextPage" not in res.json():
+                error_message = f"Get {type} nextPage keyword is missing in API response, url ::: {url}, response : {res}, response_json ::: {res.json()},"
+                logger.error('Get %s nextPage keyword is missing in API response, url ::: %s, response ::: %s, response_json ::: %s', type, url, res, res.json())
+                raise Exception(error_message)
+
+            nextPage=res.json()['nextPage']
+            page_no+=1
 
-            if res == None or not res.ok:
-                logger.error('Get %s API is failed, url %s', type, url)
-                return resp
-            elif "results" not in res.json() or len(res.json()['results'])==0 :
-                logger.error('Get %s API results are empty, url is %s', type, url)
-                return None
-            else:   
-                result = res.json()['results']
-                resp.append(result)
-        
-                if "nextPage" not in res.json():
-                    resp = []
-                    logger.error('Get %s nextPage keyword is missing in url %s', type, url)
-                    resp = [item for sublist in resp for item in sublist]
-                    return resp
-        
-                nextPage=res.json()['nextPage']
-                page_no+=1
-        
         resp = [item for sublist in resp for item in sublist] # To eliminate list of list (ex: [[data:{}]] -> [data:{}])
     except Exception as e:
         traceback.print_exc()
+        logger.error(f'get_paginated_api_results : Error while fetching apis, Error Cause is ::: {e}')
+        raise Exception(f'{error_message}')
     return resp
 
 
@@ -890,6 +996,40 @@ def get_app_logo(id, context, parameters):
     return get_logo_path_url(id)
 
 
+def generate_pie_chart_colors(num_colors):
+    colors=["#0077C8", "#00A3E0", "#673AB7", "#9C27B0", "#E91E63", "#F47925"]
+    random.seed(123)  # 123 Fixed seed value
+    #colors = []
+    for _ in range(num_colors):
+        hex_color = '#{:06x}'.format(random.randint(0, 0xFFFFFF))
+        colors.append(hex_color)
+    return colors
+
+
+def convert_datetime_tz_to_tz(date, format, fromTz, toTz):
+    if date is None or date == '-' or format is None:
+        return '-'
+    if fromTz is not None:
+        date = date.astimezone(fromTz)
+    if toTz is None or len(toTz) <= 0:
+        toTz = 'UTC'
+    if toTz is not None:
+        date_obj = datetime.strptime(date, "%Y-%m-%dT%H:%M:%S%z")
+        date = date_obj.astimezone(pytz.timezone(toTz))
+        date = date.strftime(format)
+        return date
+    return '-'
+
+
+def convert_date_time_with_format(date, format):
+    try:
+        date_obj = datetime.strptime(date, '%Y-%m-%dT%H:%M:%S%z')
+        formatted_date_str = date_obj.strftime(format)
+        return formatted_date_str
+    except Exception as e:
+        return '-'
+
+
 def update_run_progress(tenant_id, run_id, percent, message):
     url = BASE_API_URL + f'/reporting/api/v3/tenants/{tenant_id}/runs/{run_id}/progress'
     data = {
@@ -902,4 +1042,120 @@ def update_run_progress(tenant_id, run_id, percent, message):
     else:
         logger.error('Update run progress result updated upto %d',  percent)
         return res
-    #TODO : Implement logic here to update the run
\ No newline at end of file
+    #TODO : Implement logic here to update the run
+    
+       
+# Report Builder Methods
+def generate_dashboard_pdf(dashboard_id):
+    logger.info(f'{dashboard_id} :: Entered into dashboard pdf generation process')
+    try:
+        url = os.getenv("DASH_BOARD_PDF_SERVICE")
+        jwt_token = os.getenv("OPSRAMP_JWT_TOKEN")
+        current_date = datetime.now()
+        # file_name = APP_ID.lower() + '-' + pdf.analysis_run[:8] + '-' + current_date.strftime("%Y-%m-%d-%H-%M-%S") + '.pdf'
+        report_path = ''
+        if storage_name == 's3':
+            report_path = PLATFORM_ROUTE
+        elif storage_name == 'ceph':
+            report_path = APP_ID.lower()
+        pdf = PDF(dashboard_id, url, report_path, current_date.strftime("%Y-%m-%d-%H-%M-%S"))
+        file_name = pdf.prepare_file_name(APP_ID.lower(), 'pdf')
+        file_path = pdf.report_path + '/' + pdf.analysis_run + '/pdf/' + file_name
+        data = {
+            'domain': BASE_API_URL,
+            'report': PLATFORM_ROUTE,
+            'run': pdf.analysis_run,
+            'route': '/full-view',
+            'token': get_token(),
+            'app_id': APP_ID,
+            'size': 'A4',
+            'storage': storage_name,
+            'file_name': file_name,
+            'file_path': file_path,
+            'dashboard_id': f'/{pdf.analysis_run}'
+        }
+
+        gen_retry = 1
+        while gen_retry <= 2:
+            logging.info(f'{pdf.analysis_run} :: pdf generation trying {gen_retry} time..')
+            gen_retry += 1
+            logging.info(f'b4 generation >> full file path : {file_path}')
+            response = pdf.generate(data)
+            logging.info(f'after generation >> full file path : {file_path}')
+            if response == 504:
+                storage_retry = 1
+                file_found = False
+                while storage_retry <= API_RETRY:
+                    logging.info(f'{pdf.analysis_run} checking the file is existing in storage or not {storage_retry} time..')
+                    file_found = is_file_in_storage(file_path)
+                    if file_found == True:
+                        logging.info(f'{pdf.analysis_run} the pdf file is found in storage')
+                        return file_path
+                    else:
+                        time.sleep(storage_retry * 30)
+                        storage_retry += 1
+                        logging.info(f'{pdf.analysis_run} pdf not found in storage trying for {storage_retry} time..')
+                        continue
+            else:
+                return response
+        raise Exception(f'Generate_pdf:: pdf generation failed after max tries ({API_RETRY}), for run ::: {pdf.analysis_run}')
+    except Exception as e:
+        traceback.print_exc()
+        err_msg = f'Generate_pdf:: Exeception - pdf generation failed after max tries({API_RETRY}), for run ::: {pdf.analysis_run}'
+        raise Exception(err_msg)
+
+
+def dashboard_init_mail(org_id, dashboard_id, toEmails, filePath, fileName):
+    logger.info('Entered into init_mail method')
+    try:
+        # /tenants/{tenantId}/runs/{id}/sends
+        url = BASE_API_URL + f'/reporting/api/v3/tenants/{org_id}/reportbuilder/sendmail'
+        data={
+                "recipients":toEmails,
+                "filePath":filePath,
+                "fileName":fileName
+            }
+        t1 = int(time.time())
+        res = call_post_requests(url , data=json.dumps(data), verify=False)
+        t2 = int(time.time())
+        duration = t2-t1
+        if duration > API_TIME_STACKS:
+            logging.info('Status update response took %d (greater than %d) seconds', duration, API_TIME_STACKS)
+        logger.info('Status update response is %s', res)
+    except Exception as e:
+        logger.error("Exception raised due to : " + repr(e))
+        traceback.print_exc()
+    logger.info('finished init_mail method')
+    
+    
+def get_dashboard_result(dashboard_Id, path, field=None, default_value=None):
+    try:
+        # run_id= f'{PLATFORM_ROUTE}/{run_id}/json/{run_id}'
+        #PLATFORM_ROUTE = 'report-builder'
+        #print('pathhh', path)
+        if storage_name == 's3':
+            run_id= f'{PLATFORM_ROUTE}/{dashboard_Id}/json/{path}/{dashboard_Id}'
+            print('locationnn',run_id)
+            s3 = get_s3_client()
+            res_object = s3.get_object(Bucket=BUCKET_NAME,Key=run_id)
+            serializedObject = res_object['Body'].read()
+            result = json.loads(serializedObject)
+            if field:
+                result = result.get(field, default_value)
+            return result
+        elif storage_name == 'ceph':
+            run_id= f'{APP_ID.lower()}/{dashboard_Id}/json/{path}/{dashboard_Id}'
+            s3 = get_ceph_resource()
+            data = BytesIO()
+            res_object = s3.Bucket(BUCKET_NAME).download_fileobj(Fileobj=data,Key=run_id)
+            res_object = data.getvalue()
+            result = json.loads(res_object)
+            if field:
+                result = result.get(field, default_value)
+            return result
+        else:
+            logger.info("No storages are found")
+    except Exception:
+        logger.error('An error occurred (NoSuchKey) when calling the GetObject operation: The specified key does not exist')
+        pass
+
